# functions to calculate the statistics from the Gaussian process
# fit to the data
# and getfitnesspenalty

import pandas as pd
import numpy as np
from scipy.signal import find_peaks
from scipy import integrate
from scipy.interpolate import interp1d
import om_code.omgenutils as gu


def statserr(d):
    """
    Errors in statistics calculated from samples from a
    Gaussian process are half the interquartile range (consistent with
    fitderiv).
    """
    return gu.findiqr(d) / 2

###


def cleansc(self):
    """
    Ensure that NaNs do not change numeric variables from being floats.
    """
    floatvars = [
        "log2 OD ratio",
        "log2 OD ratio err",
        "local max gr",
        "local max gr err",
        "time of local max gr",
        "time of local max gr err",
        "area under gr vs OD",
        "area under gr vs OD err",
        "normalized area under gr vs OD",
        "normalized area under gr vs OD err",
        "area under OD",
        "area under OD err",
        "normalized area under OD",
        "normalized area under OD err",
        "OD logmaxlike",
        "max gr",
        "max gr err",
        "time of max gr",
        "time of max gr err",
        "doubling time",
        "doubling time err",
        "max OD",
        "max OD err",
        "lag time",
        "lag time err",
    ]
    for var in floatvars:
        if var in self.sc.columns:
            self.sc[var] = self.sc[var].astype(float)

###


def findsummarystats(
    dtype,
    derivname,
    logs,
    nosamples,
    f,
    t,
    e,
    c,
    s,
    findareas,
    figs,
    plotlocalmax,
    axgr,
    showpeakproperties,
    **kwargs,
):
    """
    Finds summary statistics from GP fit to time series of dtype
    """
    warning = None
    odtype = "log(" + dtype + ")" if logs else dtype
    outdf = pd.DataFrame(
        {
            "experiment": e,
            "condition": c,
            "strain": s,
            "time": t,
            "f" + odtype: f.f,
            "f" + odtype + " err": np.sqrt(f.fvar),
            derivname: f.df,
            derivname + " err": np.sqrt(f.dfvar),
            "d/dt " + derivname: f.ddf,
            "d/dt " + derivname + " err": np.sqrt(f.ddfvar),
        }
    )
    # check derivative has been sensibly defined
    if (
        np.max(np.abs(f.df)) < 1.0e-20
        and np.max(np.abs(np.diff(f.dfvar))) < 1.0e-20
    ):
        warning = (
            "\nWarning: getstats may have failed for "
            + e
            + ": "
            + s
            + " in "
            + c
        )
    # find summary statistics
    fs, gs, hs = f.fitderivsample(nosamples)
    # find local maximal derivative
    da, dt = findlocalmaxderiv(
        f, gs, axgr, figs, plotlocalmax, showpeakproperties, **kwargs
    )
    # find area under df vs f and area under f vs t
    if findareas:
        adff, andff, aft, anft = findareasunder(t, fs, gs, logs)
    else:
        adff, andff, aft, anft = np.nan, np.nan, np.nan, np.nan
    # store results
    statsdict = {
        "experiment": e,
        "condition": c,
        "strain": s,
        "local max " + derivname: np.median(da),
        "local max " + derivname + " err": statserr(da),
        "time of local max " + derivname: np.median(dt),
        "time of local max " + derivname + " err": statserr(dt),
        "area under " + derivname + " vs " + dtype: np.median(adff),
        "area under " + derivname + " vs " + dtype + " err": statserr(adff),
        "normalized area under " + derivname + " vs " + dtype: np.median(andff),
        "normalized area under " + derivname + " vs " + dtype + " err": statserr(andff),
        "area under " + dtype: np.median(aft),
        "area under " + dtype + " err": statserr(aft),
        "normalized area under " + dtype: np.median(anft),
        "normalized area under " + dtype + " err": statserr(anft),
    }
    return outdf, statsdict, warning

###


def findlocalmaxderiv(
    f, gs, axgr, figs, plotlocalmax, showpeakproperties, **kwargs
):
    """
    Check the derivative for local maxima. If so, find the local maximum
    with the highest derivative using samples, gs, of df. The keyword variables
    kwargs are passed to scipy's find_peaks.
    """
    # find peaks in mean df
    lpksmn, lpksmndict = find_peaks(f.df, **kwargs)
    if np.any(lpksmn):
        if showpeakproperties:
            # display properties of peaks
            print("Peak properties\n---")
            for prop in lpksmndict:
                print("{:15s}".format(prop), lpksmndict[prop])
        # da: samples of local max df
        # dt: samples of time of local max df
        da, dt = [], []
        # find peaks of sampled df
        for gsample in np.transpose(gs):
            tpks = find_peaks(gsample, **kwargs)[0]
            if np.any(tpks):
                da.append(np.max(gsample[tpks]))
                dt.append(f.t[tpks[np.argmax(gsample[tpks])]])
        if figs and plotlocalmax:
            # plot local max df as a point
            axgr.plot(
                np.median(dt),
                np.median(da),
                "o",
                color="yellow",
                markeredgecolor="k",
            )
        return da, dt
    else:
        # mean df has no peaks
        return np.nan, np.nan

###


def findareasunder(t, fs, gs, logs):
    """
    Given samples of f, as arguments fs, and of df, as arguments gs,
    either find the area under df/dt vs f and the area under f vs t
    or if logs find the area under d/dt log(f) vs f and the area
    under f vs t.
    """
    # adff: samples of area under df vs f
    # andff: samples of normalised area under df vs f
    # aft: samples of area under f vs t
    # anft: samples of normalised area under f vs t
    adff, andff, aft, anft = [], [], [], []
    for fsample, gsample in zip(np.transpose(fs), np.transpose(gs)):
        sf = np.exp(fsample) if logs else fsample
        # area under df vs f: integrand has f as x and df as y
        def integrand(x): return interp1d(sf, gsample)(x)
        iresult = integrate.quad(
            integrand, np.min(sf), np.max(sf), limit=100, full_output=1
        )[0]
        adff.append(iresult)
        andff.append(iresult / (np.max(sf) - np.min(sf)))
        # area under f vs t: integrand has t as x and f as y
        def integrand(x): return interp1d(t, sf)(x)
        iresult = integrate.quad(
            integrand, np.min(t), np.max(t), limit=100, full_output=1
        )[0]
        aft.append(iresult)
        anft.append(iresult / (np.max(t) - np.min(t)))
    return adff, andff, aft, anft



# functions for taking subsets of the data
import numpy as np
import om_code.omgenutils as gu
import om_code.omerrors as errors


def getsubset(
    inst,
    type,
    set="all",
    includes=False,
    excludes=False,
    nonull=False,
    nomedia=False,
):
    """
    Returns a subset of either the conditions or strains.

    Parameters
    --
    inst: instance of platereader
    type: string
        Either 'c' (conditions) or 's' (strains).
    set: list of strings
        List of items to include (default is 'all').
    includes: string
        Select only items with this string in their name.
    excludes: string
        Ignore any items with this string in their name.
    nonull: boolean
        If True, ignore Null strain.
    nomedia: boolean
        If True, ignores 'media' condition.

    Returns
    -------
    sset: list of strings
    """
    if set == "all" or includes or excludes:
        if type == "c":
            sset = list(
                np.unique(
                    [
                        con
                        for e in inst.allconditions
                        for con in inst.allconditions[e]
                    ]
                )
            )
            if nomedia and "media" in sset:
                sset.pop(sset.index("media"))
        elif type == "s":
            sset = list(
                np.unique(
                    [
                        str
                        for e in inst.allstrains
                        for str in inst.allstrains[e]
                    ]
                )
            )
            if nonull and "Null" in sset:
                sset.pop(sset.index("Null"))
        else:
            sset = inst.allexperiments
        # find those items containing keywords given in 'includes'
        if includes:
            includes = gu.makelist(includes)
            newset = []
            for s in sset:
                gotone = 0
                for item in includes:
                    if item in s:
                        gotone += 1
                if gotone == len(includes):
                    newset.append(s)
            sset = newset
        # remove any items containing keywords given in 'excludes'
        if excludes:
            excludes = gu.makelist(excludes)
            exs = []
            for s in sset:
                for item in excludes:
                    if item in s:
                        exs.append(s)
                        break
            for ex in exs:
                sset.pop(sset.index(ex))
    else:
        sset = gu.makelist(set)
    if sset:
        # sort by numeric values in list entries
        return sorted(sset, key=gu.natural_keys)
    else:
        if includes:
            raise errors.getsubset(
                "Nothing found for " + " and ".join(includes)
            )
        else:
            raise errors.getsubset("Nothing found")


###


def getexps(inst, experiments, experimentincludes, experimentexcludes):
    """
    Returns list of experiments
    """
    if experimentincludes or experimentexcludes:
        exps = getsubset(
            inst,
            "e",
            includes=experimentincludes,
            excludes=experimentexcludes,
        )
    elif experiments == "all":
        exps = inst.allexperiments
    else:
        exps = gu.makelist(experiments)
    return exps


###


def getcons(inst, conditions, conditionincludes, conditionexcludes, nomedia):
    """
    Returns list of conditions
    """
    if conditionincludes or conditionexcludes:
        cons = getsubset(
            inst,
            "c",
            includes=conditionincludes,
            excludes=conditionexcludes,
            nomedia=nomedia,
        )
    elif conditions == "all":
        cons = list(
            np.unique(
                [
                    con
                    for e in inst.allconditions
                    for con in inst.allconditions[e]
                ]
            )
        )
        if nomedia and "media" in cons:
            cons.pop(cons.index("media"))
    else:
        cons = gu.makelist(conditions)
    return sorted(cons, key=gu.natural_keys)


###


def getstrs(inst, strains, strainincludes, strainexcludes, nonull):
    """
    Returns list of strains
    """
    if strainincludes or strainexcludes:
        strs = getsubset(
            inst,
            "s",
            includes=strainincludes,
            excludes=strainexcludes,
            nonull=nonull,
        )
    elif strains == "all":
        strs = list(
            np.unique(
                [str for e in inst.allstrains for str in inst.allstrains[e]]
            )
        )
        if nonull and "Null" in strs:
            strs.pop(strs.index("Null"))
    else:
        strs = gu.makelist(strains)
    if nonull and "Null" in strs:
        strs.pop(strs.index("Null"))
    return sorted(strs, key=gu.natural_keys)


###


def getall(
    inst,
    experiments,
    experimentincludes,
    experimentexcludes,
    conditions,
    conditionincludes,
    conditionexcludes,
    strains,
    strainincludes,
    strainexcludes,
    nonull=True,
    nomedia=True,
):
    """
    Returns experiments, conditions, and strains
    """
    exps = getexps(inst, experiments, experimentincludes, experimentexcludes)
    cons = getcons(
        inst, conditions, conditionincludes, conditionexcludes, nomedia
    )
    strs = getstrs(inst, strains, strainincludes, strainexcludes, nonull)
    return exps, cons, strs


###


def extractwells(r_df, s_df, experiment, condition, strain, datatypes):
    """
    Extracts a list of matrices for each dtype in
    datatypes for the given experiment, condition, and strain with each
    column in each matrix having the data for one well
    """
    datatypes = gu.makelist(datatypes)
    # restrict time if necessary
    lrdf = r_df[
        (r_df.time >= s_df.time.min()) & (r_df.time <= s_df.time.max())
    ]
    # extract data
    df = lrdf.query(
        "experiment == @experiment and condition == @condition "
        "and strain == @strain"
    )
    matrices = []
    for dtype in datatypes:
        df2 = df[[dtype, "well"]]
        df2well = df2.groupby("well")[dtype].apply(list)
        matrices.append(np.transpose([df2well[w] for w in df2well.index]))
    if len(datatypes) == 1:
        # return array
        return matrices[0]
    else:
        # return list of arrays
        return matrices



# functions for correcting for non-linearities in the OD, for
# the fluorescence of the media, and for autofluorescence
import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import interp1d
import gaussianprocessderivatives as gp
from statsmodels.nonparametric.smoothers_lowess import lowess
from scipy.optimize import minimize_scalar
import pandas as pd
import om_code.sunder as sunder
import om_code.omerrors as errors
import om_code.omgenutils as gu


def findODcorrection(wdirpath, ODfname, figs, odmatch, cvfn, bd, gp_results):
    """
    Uses a Gaussian process to fit serial dilution data
    to correct for non-linearities in the relationship between OD and cell
    density. The data are either loaded from file ODfname or the default
    data for haploid yeast growing in glucose are used (collected by
    L Bandiera).
    """
    print("Fitting dilution data for OD correction for non-linearities")
    if ODfname is not None:
        try:
            od_df = pd.read_csv(
                str(wdirpath / ODfname), sep=None, engine="python", header=None
            )
            print("Using", ODfname)
            od_data = od_df.to_numpy()
            od, dilfac = od_data[:, 0], od_data[:, 1]
        except (FileNotFoundError, OSError):
            raise errors.FileNotFound(str(wdirpath / ODfname))
    else:
        print("Using default data for haploid yeast in glucose")
        # Lucia's data
        od = np.array(
            [
                1.38192778,
                1.15388333,
                0.96450556,
                0.78569444,
                0.61685,
                0.45291666,
                0.33923888,
                0.24501666,
                0.18033889,
                0.12463889,
                0.08463889,
                0.05783889,
                0.04019444,
                0.01776111,
                0.01790556,
                0.66888333,
                0.47343889,
                0.35310556,
                0.25619444,
                0.18245,
                0.19291667,
                0.12428333,
                0.08719444,
                0.05910555,
                0.04203889,
                0.02910556,
                0.02035,
                0.01392778,
                0.00946111,
                0.00651666,
                1.06491665,
                0.85690557,
                0.71552777,
                0.52646112,
                0.39376111,
                0.28846111,
                0.20293889,
                0.14296111,
                0.10538333,
                0.07017222,
                0.04939444,
                0.03397222,
                0.02639445,
                0.01780556,
                0.01463889,
                0.34130555,
                0.21385,
                0.15140556,
                0.10612778,
                0.07330556,
                0.05147222,
                0.03702778,
                0.02599445,
                0.01739444,
                0.01188333,
                0.00956111,
                0.00572778,
                0.00491667,
                0.00275,
                0.00316111,
            ]
        )
        dilfac = np.array(
            [
                1.0000000e00,
                7.1968567e-01,
                5.1794747e-01,
                3.7275937e-01,
                2.6826958e-01,
                1.9306977e-01,
                1.3894955e-01,
                1.0000000e-01,
                7.1968570e-02,
                5.1794750e-02,
                3.7275940e-02,
                2.6826960e-02,
                1.9306980e-02,
                1.3894950e-02,
                1.0000000e-02,
                3.0484230e-01,
                2.1939064e-01,
                1.5789230e-01,
                1.1363283e-01,
                8.1779920e-02,
                5.8855830e-02,
                4.2357700e-02,
                3.0484230e-02,
                2.1939060e-02,
                1.5789230e-02,
                1.1363280e-02,
                8.1779900e-03,
                5.8855800e-03,
                4.2357700e-03,
                3.0484200e-03,
                4.4830728e-01,
                3.2264033e-01,
                2.3219962e-01,
                1.6711074e-01,
                1.2026721e-01,
                8.6554590e-02,
                6.2292100e-02,
                4.4830730e-02,
                3.2264030e-02,
                2.3219960e-02,
                1.6711070e-02,
                1.2026720e-02,
                8.6554600e-03,
                6.2292100e-03,
                4.4830700e-03,
                7.7885310e-02,
                5.6052940e-02,
                4.0340500e-02,
                2.9032480e-02,
                2.0894260e-02,
                1.5037300e-02,
                1.0822130e-02,
                7.7885300e-03,
                5.6052900e-03,
                4.0340500e-03,
                2.9032500e-03,
                2.0894300e-03,
                1.5037300e-03,
                1.0822100e-03,
                7.7885000e-04,
            ]
        )
    # process data
    dilfac = dilfac[np.argsort(od)]
    od = np.sort(od)
    if odmatch is not None:
        # rescale so that OD and dilfac match at a particular OD
        # compares better with Warringer & Blomberg, Yeast 2003,
        # and rescaled OD is larger
        dilfacmatch = interp1d(od, dilfac)(odmatch)
        y = dilfac / dilfacmatch * odmatch
    else:
        y = dilfac
    # set up Gaussian process
    bddict = {
        "sqexplin": {0: (-4, 2), 1: (-3, 1), 2: (-6, 1), 3: (-6, 1)},
        "nn": {0: (-4, 2), 1: (-3, 1), 2: (-6, 1)},
        "sqexp": {0: (-4, 2), 1: (-3, 1), 2: (-6, 1)},
        "matern": {0: (-4, 2), 1: (-3, 1), 2: (-6, 1)},
    }
    # find bounds
    if bd:
        bds = gu.mergedicts(original=bddict[cvfn], update=bd)
    else:
        bds = bddict[cvfn]
    gc = getattr(gp, cvfn + "GP")(bds, od, y)
    # run Gaussian process
    gc.findhyperparameters(noruns=5, exitearly=True, quiet=True)
    if gp_results:
        gc.results()
    gc.predict(od)
    if figs:
        plt.figure()
        gc.sketch(".")
        plt.grid(True)
        plt.xlabel("OD")
        plt.ylabel("corrected OD (relative cell numbers)")
        if ODfname:
            plt.title("Fitting " + ODfname)
        else:
            plt.title("for haploid budding yeast in glucose")
        plt.show()
    return gc


###


def performmediacorrection(r_df, dtype, exp, condition, figs, commonmedia, frac):
    """
    Uses lowess to smooth the media data from Null
    wells over time and subtracts the smoothed values from the data.
    """
    # find data for correction with condition equal to commonmedia
    df = r_df.query(
        "experiment == @exp and condition == @commonmedia" " and strain == 'Null'"
    )
    if df.empty:
        # no data
        print(
            ' No well annotated "Null" was found for',
            commonmedia,
            "in experiment",
            exp,
        )
        print(" Correcting for media for", dtype, "in", commonmedia, "abandoned!")
        return False, None
    else:
        # there is data - change r dataframe
        rtest = (r_df.experiment == exp) & (r_df.condition == condition)
        t, data = df["time"].to_numpy(), df[dtype].to_numpy()
        # find correction
        res = lowess(data, t, frac=frac)
        correctionfn = interp1d(
            res[:, 0],
            res[:, 1],
            fill_value=(res[0, 1], res[-1, 1]),
            bounds_error=False,
        )
        if figs:
            plt.figure()
            plt.plot(t, data, "ro", res[:, 0], res[:, 1], "b-")
            plt.xlabel("time (hours)")
            plt.title(exp + ": media correction for " + dtype + " in " + condition)
            plt.show()
        # perform correction
        r_df.loc[rtest, dtype] = r_df[rtest][dtype] - correctionfn(r_df[rtest]["time"])
        # check for any negative values
        negvalues = ""
        for s in np.unique(r_df[rtest]["strain"][r_df[rtest][dtype] < 0]):
            if s != "Null":
                wstr = "\t" + dtype + ": " + s + " in " + condition + " for wells "
                for well in np.unique(
                    r_df[rtest][r_df[rtest].strain == s]["well"][r_df[rtest][dtype] < 0]
                ):
                    wstr += well + " "
                wstr += "\n"
                negvalues += wstr
        return True, negvalues


###


def correctauto1(
    self,
    f,
    refstrain,
    figs,
    experiments,
    experimentincludes,
    experimentexcludes,
    conditions,
    conditionincludes,
    conditionexcludes,
    strains,
    strainincludes,
    strainexcludes,
):
    """
    Corrects for autofluorescence for experiments with measured
    emissions at one wavelength using the fluorescence of the reference strain
    interpolated to the OD of the tagged strain. This in principle corrects too
    for the fluorescence of the medium, although running correctmedia is still
    preferred.
    """
    print("Correcting autofluorescence using", f[0])
    for e in sunder.getexps(self, experiments, experimentincludes, experimentexcludes):
        for c in sunder.getcons(
            self,
            conditions,
            conditionincludes,
            conditionexcludes,
            nomedia=True,
        ):
            # process reference strain
            refstrfn = processref1(self, f, refstrain, figs, e, c)
            # correct strains
            for s in sunder.getstrs(
                self, strains, strainincludes, strainexcludes, nonull=True
            ):
                if not self.sc[
                    (self.sc.experiment == e)
                    & (self.sc.condition == c)
                    & (self.sc.strain == s)
                ][f[0] + " corrected for autofluorescence"].any():
                    od, rawfl = sunder.extractwells(
                        self.r, self.s, e, c, s, ["OD", f[0]]
                    )
                    # no data
                    if od.size == 0 or rawfl.size == 0:
                        continue
                    # correct autofluorescence for each replicate
                    fl = np.transpose(
                        [rawfl[:, i] - refstrfn(od[:, i]) for i in range(od.shape[1])]
                    )
                    flperod = np.transpose(
                        [
                            (rawfl[:, i] - refstrfn(od[:, i])) / od[:, i]
                            for i in range(od.shape[1])
                        ]
                    )
                    # replace negative values with NaNs
                    fl[fl < 0] = np.nan
                    flperod[flperod < 0] = np.nan
                    nonans = np.count_nonzero(np.isnan(fl))
                    if np.any(nonans):
                        print(
                            "Warning -",
                            e + ":",
                            s,
                            "in",
                            c,
                            "\n",
                            nonans,
                            "corrected data points are"
                            " NaN because the corrected fluorescence"
                            " was negative",
                        )
                        if nonans == fl.size:
                            print(
                                "Warning -",
                                e + ":",
                                s,
                                "in",
                                c,
                                "\n",
                                "Corrected fluorescence is all NaN",
                            )
                    # store results
                    bname = "c-" + f[0]
                    autofdict = {
                        "experiment": e,
                        "condition": c,
                        "strain": s,
                        "time": self.s.query(
                            "experiment == @e and condition == @c " "and strain == @s"
                        )["time"].to_numpy(),
                        bname: np.nanmean(fl, 1),
                        bname + " err": nanstdzeros2nan(fl, 1),
                        bname + "perOD": np.nanmean(flperod, 1),
                        bname + "perOD err": nanstdzeros2nan(flperod, 1),
                    }
                    autofdf = pd.DataFrame(autofdict)
                    if bname not in self.s.columns:
                        # extend dataframe
                        self.s = pd.merge(self.s, autofdf, how="outer")
                    else:
                        # update dataframe
                        self.s = gu.absorbdf(
                            self.s,
                            autofdf,
                            ["experiment", "condition", "strain", "time"],
                        )
                    # record that correction has occurred
                    self.sc.loc[
                        (self.sc.experiment == e)
                        & (self.sc.condition == c)
                        & (self.sc.strain == s),
                        f[0] + " corrected for autofluorescence",
                    ] = True


###


def processref1(self, f, refstrain, figs, experiment, condition):
    """
    Processes reference strain for data with one fluorescence
    measurement. Uses lowess to smooth the fluorescence of the reference
    strain as a function of OD.

    Parameters
    ----------
    f: string
        The fluorescence to be corrected. For example, ['mCherry'].
    refstrain: string
        The reference strain. For example, 'WT'.
    figs: boolean
        If True, display fits of the reference strain's fluorescence.
    experiment: string
        The experiment to be corrected.
    condition: string
        The condition to be corrected.

    Returns
    -------
    refstrfn: function
        The reference strain's fluorescence as a function of OD.
    """
    e, c = experiment, condition
    print(
        e + ": Processing reference strain",
        refstrain,
        "for",
        f[0],
        "in",
        c,
    )
    od, fl = sunder.extractwells(self.r, self.s, e, c, refstrain, ["OD", f[0]])
    if od.size == 0 or fl.size == 0:
        raise errors.CorrectAuto(e + ": " + refstrain + " not found in " + c)
    else:
        odf = od.flatten("F")
        flf = fl.flatten("F")
        # smooth fluorescence as a function of OD using lowess to minimize
        # refstrain's autofluorescence

        def choosefrac(frac):
            res = lowess(flf, odf, frac=frac)
            refstrfn = interp1d(
                res[:, 0],
                res[:, 1],
                fill_value=(res[0, 1], res[-1, 1]),
                bounds_error=False,
            )
            # max gives smoother fits than mean
            return np.max(np.abs(flf - refstrfn(odf)))

        res = minimize_scalar(choosefrac, bounds=(0.1, 0.99), method="bounded")
        # choose the optimum frac
        frac = res.x if res.success else 0.33
        res = lowess(flf, odf, frac=frac)
        refstrfn = interp1d(
            res[:, 0],
            res[:, 1],
            fill_value=(res[0, 1], res[-1, 1]),
            bounds_error=False,
        )
        if figs:
            # plot fit
            plt.figure()
            plt.plot(odf, flf, ".", alpha=0.5)
            plt.plot(res[:, 0], res[:, 1])
            plt.xlabel("OD")
            plt.ylabel(f[0])
            plt.title(e + ": " + refstrain + " for " + c)
            plt.show()
        return refstrfn


###


def correctauto2(
    self,
    f,
    refstrain,
    figs,
    experiments,
    experimentincludes,
    experimentexcludes,
    conditions,
    conditionincludes,
    conditionexcludes,
    strains,
    strainincludes,
    strainexcludes,
):
    """
    Corrects for autofluorescence using spectral unmixing
    for experiments with measured emissions at two wavelengths.

    References
    ----------
    CA Lichten, R White, IB Clark, PS Swain (2014). Unmixing of fluorescence
    spectra to resolve quantitative time-series measurements of gene
    expression in plate readers.
    BMC Biotech, 14, 1-11.
    """
    # correct for autofluorescence
    print("Correcting autofluorescence using", f[0], "and", f[1])
    for e in sunder.getexps(self, experiments, experimentincludes, experimentexcludes):
        for c in sunder.getcons(
            self,
            conditions,
            conditionincludes,
            conditionexcludes,
            nomedia=True,
        ):
            # process reference strain
            refqrfn = processref2(self, f, refstrain, figs, e, c)
            # process other strains
            for s in sunder.getstrs(
                self, strains, strainincludes, strainexcludes, nonull=True
            ):
                if s != refstrain and not (
                    self.sc[
                        (self.sc.experiment == e)
                        & (self.sc.condition == c)
                        & (self.sc.strain == s)
                    ][f[0] + " corrected for autofluorescence"].any()
                ):
                    f0, f1 = sunder.extractwells(self.r, self.s, e, c, s, f)
                    if f0.size == 0 or f1.size == 0:
                        continue
                    nodata, nr = f0.shape
                    # set negative values to NaNs
                    f0[f0 < 0] = np.nan
                    f1[f1 < 0] = np.nan
                    # use mean OD for correction
                    odmean = self.s.query(
                        "experiment == @e and condition == @c and strain == @s"
                    )["OD mean"].to_numpy()
                    # remove autofluorescence
                    ra = refqrfn(odmean)
                    fl = applyautoflcorrection(self, ra, f0, f1)
                    od = sunder.extractwells(self.r, self.s, e, c, s, "OD")
                    flperod = fl / od
                    # set negative values to NaNs
                    fl[fl < 0] = np.nan
                    flperod[flperod < 0] = np.nan
                    # store corrected fluorescence
                    bname = "c-" + f[0]
                    autofdict = {
                        "experiment": e,
                        "condition": c,
                        "strain": s,
                        "time": self.s.query(
                            "experiment == @e and condition == @c" " and strain == @s"
                        )["time"].to_numpy(),
                        bname: np.nanmean(fl, 1),
                        bname + " err": nanstdzeros2nan(fl, 1),
                        bname + "perOD": np.nanmean(flperod, 1),
                        bname + "perOD err": nanstdzeros2nan(flperod, 1),
                    }
                    # add to dataframe
                    self.s = gu.absorbdf(
                        self.s,
                        pd.DataFrame(autofdict),
                        ["experiment", "condition", "strain", "time"],
                    )
                    self.sc.loc[
                        (self.sc.experiment == e)
                        & (self.sc.condition == c)
                        & (self.sc.strain == s),
                        f[0] + " corrected for autofluorescence",
                    ] = True


###


def processref2(self, f, refstrain, figs, experiment, condition):
    """
    Processes reference strain data for spectral unmixing
    (for experiments with two fluorescence measurements). Uses lowess to
    smooth the ratio of emitted fluorescence measurements so that the
    reference strain's data is corrected to zero as best as possible.

    Parameters
    ----------
    f: list of strings
        The fluorescence measurements. For example, ['GFP', 'AutoFL'].
    refstrain: string
        The reference strain. For example, 'WT'.
    figs: boolean
        If True, display fits of the fluorescence ratios.
    experiment: string
        The experiment to be corrected.
    condition: string
        The condition to be corrected.

    Returns
    -------
    qrfn: function
        The ratio of the two fluorescences for the reference strain as a
        function of OD.
    """
    e, c = experiment, condition
    print(
        e + ": Processing reference strain",
        refstrain,
        "for",
        f[0],
        "in",
        c,
    )
    # refstrain data
    f0, f1, od = sunder.extractwells(self.r, self.s, e, c, refstrain, f + ["OD"])
    if f0.size == 0 or f1.size == 0 or od.size == 0:
        raise errors.CorrectAuto(e + ": " + refstrain + " not found in " + c)
    else:
        f0[f0 < 0] = np.nan
        f1[f1 < 0] = np.nan
        odf = od.flatten("F")
        odrefmean = np.mean(od, 1)
        qrf = (f1 / f0).flatten("F")
        if np.all(np.isnan(qrf)):
            raise errors.CorrectAuto(
                e + ": " + refstrain + " in " + c + " has too many NaNs"
            )
        # smooth to minimize autofluorescence in refstrain

        def choosefrac(frac):
            res = lowess(qrf, odf, frac)
            qrfn = interp1d(
                res[:, 0],
                res[:, 1],
                fill_value=(res[0, 1], res[-1, 1]),
                bounds_error=False,
            )
            flref = applyautoflcorrection(self, qrfn(odrefmean), f0, f1)
            return np.max(np.abs(flref))

        res = minimize_scalar(choosefrac, bounds=(0.1, 0.99), method="bounded")
        # calculate the relationship between qr and OD
        frac = res.x if res.success else 0.95
        res = lowess(qrf, odf, frac)
        qrfn = interp1d(
            res[:, 0],
            res[:, 1],
            fill_value=(res[0, 1], res[-1, 1]),
            bounds_error=False,
        )
        if figs:
            plt.figure()
            plt.plot(odf, qrf, ".", alpha=0.5)
            plt.plot(res[:, 0], res[:, 1])
            plt.xlabel("OD")
            plt.ylabel(f[1] + "/" + f[0])
            plt.title(e + ": " + refstrain + " in " + c)
            plt.show()
        # check autofluorescence correction for reference strain
        flref = applyautoflcorrection(self, qrfn(odrefmean), f0, f1)
        flrefperod = flref / od
        # set negative values to NaNs
        flref[flref < 0] = np.nan
        flrefperod[flrefperod < 0] = np.nan
        # store results
        bname = "c-" + f[0]
        autofdict = {
            "experiment": e,
            "condition": c,
            "strain": refstrain,
            "time": self.s.query(
                "experiment == @e and condition == @c and strain == @refstrain"
            )["time"].to_numpy(),
            bname: np.nanmean(flref, 1),
            bname + "perOD": np.nanmean(flrefperod, 1),
            bname + " err": nanstdzeros2nan(flref, 1),
            bname + "perOD err": nanstdzeros2nan(flrefperod, 1),
        }
        if bname not in self.s.columns:
            self.s = pd.merge(self.s, pd.DataFrame(autofdict), how="outer")
        else:
            self.s = gu.absorbdf(
                self.s,
                pd.DataFrame(autofdict),
                ["experiment", "condition", "strain", "time"],
            )
        return qrfn


###


def applyautoflcorrection(self, ra, f0data, f1data):
    """
    Corrects for autofluorescence returning an array of replicates.
    """
    nr = f0data.shape[1]
    raa = np.reshape(np.tile(ra, nr), (np.size(ra), nr), order="F")
    return (raa * f0data - f1data) / (raa - self._gamma * np.ones(np.shape(raa)))


###


def nanstdzeros2nan(a, axis=None):
    """
    nanstd but setting zeros to nan
    """
    err = np.nanstd(a, axis)
    err[err == 0] = np.nan
    return err



# functions to plot the different data frames and the plate as a whole
import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import pandas as pd
import om_code.omgenutils as gu
import om_code.omerrors as errors


def plotplate(basedf, exps, dtype):
    """
    Plots the data for each well following the layout of
    a 96-well plate.

    Parameters
    --
    basedf: DataFrame
        The r dataframe.
    exps: float
        The name of the experiments.
    dtype: float
        The data type to be plotted: 'OD', 'GFP', etc.
    """
    for e in exps:
        plt.figure()
        # first create an empty plate - in case of missing wells
        ax = []
        for rowl in range(8):
            for coll in np.arange(1, 13):
                sindex = coll + 12 * rowl
                axi = plt.subplot(8, 12, sindex)
                ax.append(axi)
                plt.tick_params(labelbottom=False, labelleft=False)
                # label well locations
                for j in range(12):
                    if sindex == j + 1:
                        plt.title(j + 1)
                for j, k in enumerate(np.arange(1, 96, 12)):
                    if sindex == k:
                        plt.ylabel("ABCDEFGH"[j] + " ", rotation=0)
        # fill in the wells that have been measured
        for pl in basedf.query("experiment == @e")["well"].unique():
            rowl = "ABCDEFGH".index(pl[0])
            coll = int(pl[1:])
            sindex = coll + 12 * rowl
            wd = basedf.query("experiment == @e and well == @pl")
            ax[sindex - 1].plot(wd["time"].to_numpy(), wd[dtype].to_numpy(), "-")
        plt.suptitle(e + ": " + dtype)
        plt.show()


###


def plot_wells(
    x,
    y,
    basedf,
    exps,
    cons,
    strs,
    style="condition",
    size=None,
    kind="line",
    col=None,
    row=None,
    xlim=None,
    ylim=None,
    title=None,
    figsize=None,
    messages=False,
    **kwargs
):
    """
    Plot data from the individual wells.
    Data for each experiment, condition, and strain are plotted in
    a separate figure unless row and col are specified.
    """
    for e in exps:
        if row and col:
            # use facetgrid to show multiple plots simultaneously
            df = basedf.query(
                "experiment == @e and condition == @cons and strain == @strs"
            )
            sfig = sns.FacetGrid(df, row=row, col=col)
            for (row_var, col_var), facet_df in df.groupby([row, col]):
                ax = sfig.axes[
                    sfig.row_names.index(row_var),
                    sfig.col_names.index(col_var),
                ]
                sns.lineplot(x=x, y=y, hue="well", data=facet_df, ax=ax)
                ax.set(xlabel="", ylabel="")
            sfig.set_axis_labels(x, y)
            sfig.set_titles()
            if title:
                sfig.fig.suptitle(title)
            else:
                sfig.fig.suptitle(e)
            if xlim is not None:
                plt.xlim(xlim)
            if ylim is not None:
                plt.ylim(ylim)
            if figsize and len(figsize) == 2:
                sfig.fig.set_figwidth(figsize[0])
                sfig.fig.set_figheight(figsize[1])
            plt.tight_layout()
            plt.show()
        else:
            # create one plot for each strain and condition
            for c in cons:
                for s in strs:
                    df = basedf.query(
                        "experiment == @e and condition == @c and strain == @s"
                    )
                    if df.empty:
                        if messages:
                            print(e + ":", "No data found for", s, "in", c)
                    else:
                        sfig = sns.relplot(
                            x=x,
                            y=y,
                            data=df,
                            hue="well",
                            kind=kind,
                            style=style,
                            size=size,
                            **kwargs,
                        )
                        if title:
                            sfig.fig.suptitle(title)
                        else:
                            sfig.fig.suptitle(e + ": " + s + " in " + c)
                        if xlim is not None:
                            plt.xlim(xlim)
                        if ylim is not None:
                            plt.ylim(ylim)
                        plt.tight_layout()
                        plt.show()


###


def plot_rs(
    x,
    y,
    basedf,
    exps,
    cons,
    strs,
    hue="strain",
    style="condition",
    size=None,
    kind="line",
    col=None,
    row=None,
    height=5,
    aspect=1,
    xlim=None,
    ylim=None,
    title=None,
    figsize=None,
    sortby=False,
    returnfacetgrid=False,
    **kwargs
):
    """
    Plots time-series data from the .r or .s dataframes.
    """
    # plot time series
    df = basedf.query("experiment == @exps and condition == @cons and strain == @strs")
    if df.empty:
        print("No data found")
    else:
        if sortby:
            df = df.sort_values(by=gu.makelist(sortby))
        # add warnings for poor choice of seaborn's parameters - may cause
        # inadvertent averaging
        if hue == style:
            print(
                'Warning: "hue" and "style" have both been set to "'
                + hue
                + '" and there may be unintended averaging'
            )
        if (
            x != "commontime"
            and len(df["experiment"].unique()) > 1
            and hue != "experiment"
            and size != "experiment"
            and style != "experiment"
            and col != "experiment"
        ):
            print(
                "Warning: there are multiple experiments, but neither "
                '"hue", "style", nor "size" is set to "experiment" and there'
                " may be averaging over experiments"
            )

        if "units" not in kwargs:
            # try to augment df to allow seaborn to estimate errors
            df = augmentdf(df, y)
        # plot
        sfig = sns.relplot(
            x=x,
            y=y,
            data=df,
            hue=hue,
            kind=kind,
            style=style,
            ci="sd",
            size=size,
            col=col,
            row=row,
            aspect=aspect,
            height=height,
            **kwargs,
        )
        if title:
            sfig.fig.suptitle(title)
        if xlim is not None:
            sfig.set(xlim=xlim)
        if ylim is not None:
            sfig.set(ylim=ylim)
        if figsize and len(figsize) == 2:
            sfig.fig.set_figwidth(figsize[0])
            sfig.fig.set_figheight(figsize[1])
        plt.show()
        if returnfacetgrid:
            return sfig
        else:
            return None


###


def plot_sc(
    x,
    y,
    basedf,
    exps,
    cons,
    strs,
    hue="strain",
    style="condition",
    size=None,
    kind="line",
    col=None,
    row=None,
    height=5,
    aspect=1,
    xlim=None,
    ylim=None,
    figsize=None,
    title=None,
    sortby=False,
    **kwargs
):
    """
    Plots summary statistics from the .sc dataframe.
    """
    # plot summary stats
    df = basedf.query("experiment == @exps and condition == @cons and strain == @strs")
    xcols = df.columns[df.columns.str.startswith(x)]
    ycols = df.columns[df.columns.str.startswith(y)]
    df = df[
        np.unique(["experiment", "condition", "strain"] + list(xcols) + list(ycols))
    ].dropna()
    if df.empty:
        print("No data found")
    else:
        if sortby:
            df = df.sort_values(by=gu.makelist(sortby))
        sfig = sns.relplot(
            x=x,
            y=y,
            data=df,
            hue=hue,
            kind="scatter",
            style=style,
            size=size,
            col=col,
            row=row,
            aspect=aspect,
            height=height,
            **kwargs,
        )
        if xlim is not None:
            sfig.set(xlim=xlim)
        if ylim is not None:
            sfig.set(ylim=ylim)
        if row is None and col is None:
            # add error bars
            # find coordinates of points in relplot
            xc, yc = [], []
            for point_pair in sfig.ax.collections:
                for xp, yp in point_pair.get_offsets():
                    xc.append(xp)
                    yc.append(yp)
            # add error bars
            xerr = df[x + " err"] if x + " err" in df.columns else None
            yerr = df[y + " err"] if y + " err" in df.columns else None
            sfig.ax.errorbar(
                xc,
                yc,
                xerr=xerr,
                yerr=yerr,
                fmt=" ",
                ecolor="dimgray",
                alpha=0.5,
            )
        plt.show()


###


def plotfinddf(self, x, y):
    """
    Finds the correct dataframe for plotting y versus x

    Parameters
    ----------
    self: a platereader instance
    x: string
        Name of x-variable.
    y: string
        Name of y-variable.

    Returns
    -------
    basedf: dataframe
        The dataframe that contains the x and y variables.
    dfname: string
        The name of the dataframe.
    """
    # choose the correct dataframe
    if hasattr(self, "r") and x in self.r.columns and y in self.r.columns:
        # raw data (with wells)
        basedf = self.r
        dfname = "r"
    elif x in self.s.columns and y in self.s.columns:
        # processed data (no wells)
        basedf = self.s
        dfname = "s"
    elif x in self.sc.columns and y in self.sc.columns:
        # summary stats
        basedf = self.sc
        dfname = "sc"
    else:
        raise errors.PlotError(
            "The variables x= "
            + x
            + " and y= "
            + y
            + " cannot be plotted against each other because they are not in "
            + " the same dataframe"
        )
    return basedf, dfname


###


def augmentdf(df, datatype):
    """
    Artifically augments dataframe using 'err' (if present in the
    dataframe) to allow Seaborn to generate errors, otherwise returns
    original dataframe.

    Note we call seaborn with ci = "sd" and so use sqrt(3/2) * error
    because seaborn calculates the standard deviation from the augmented data
    (the mean, the mean + std, and the mean - std) and so gets
    std/sqrt(3/2) otherwise because there are three data points.
    """
    if datatype + " err" in df:
        derr = datatype + " err"
    elif "mean" in datatype and datatype.split(" mean")[0] + " err" in df:
        derr = datatype.split(" mean")[0] + " err"
    else:
        derr = False
        # returned if df is df_r
        return df
    if derr:
        df.insert(0, "augtype", "mean")
        mn = df[datatype].to_numpy()
        err = df[derr].to_numpy()
        # add std
        dfp = df.copy()
        dfp[datatype] = mn + np.sqrt(3 / 2) * err
        dfp["augtype"] = "+err"
        # minus std
        dfm = df.copy()
        dfm[datatype] = mn - np.sqrt(3 / 2) * err
        dfm["augtype"] = "-err"
        # concat
        df = pd.concat([df, dfp, dfm], ignore_index=True)
    return df



# General admin functions
# generating and updating dataframes and the progress dictionary
import pandas as pd


def initialiseprogress(self, experiment):
    """
    Initialises progress dictionary.
    """
    self.progress["ignoredwells"][experiment] = []
    self.progress["negativevalues"][experiment] = False


###


def makewellsdf(df_r):
    """
    Makes a dataframe that stores the contents of
    the wells
    """
    df = df_r[["experiment", "condition", "strain", "well"]].drop_duplicates()
    df = df.reset_index(drop=True)
    return df


###


def make_s(self, tmin=None, tmax=None):
    """
    Generate s datafram by calculating means and variances of all datatypes
    from raw data
    """
    # restrict time
    if tmin and not tmax:
        rdf = self.r[self.r.time >= tmin]
    elif tmax and not tmin:
        rdf = self.r[self.r.time <= tmax]
    elif tmin and tmax:
        rdf = self.r[(self.r.time >= tmin) & (self.r.time <= tmax)]
    else:
        rdf = self.r
    # find means
    df1 = (
        rdf.groupby(["experiment", "condition", "strain", "time"]).mean().reset_index()
    )
    for exp in self.allexperiments:
        for dtype in self.datatypes[exp]:
            df1 = df1.rename(columns={dtype: dtype + " mean"})
    # find errors
    df2 = rdf.groupby(["experiment", "condition", "strain", "time"]).std().reset_index()
    for exp in self.allexperiments:
        for dtype in self.datatypes[exp]:
            df2 = df2.rename(columns={dtype: dtype + " err"})
    return pd.merge(df1, df2)


###


def update_s(self):
    """
    Updates means and errors of all datatypes from raw data
    """
    # find tmin and tmax in case restrict_time has been called
    tmin = self.s.time.min()
    tmax = self.s.time.max()
    # recalculate s dataframe
    self.s = make_s(self, tmin, tmax)


###
